# -*- coding: utf-8 -*-
from tqdm import tqdm
import numpy as np
import os
import six
import pickle
import gzip
from collections import defaultdict


class Vocab(object):
    """
    Defines a vocabulary object that will be used to numericalize a field.
    """
    def __init__(self,
                 counter,
                 max_size=None,
                 min_freq=1,
                 init_token=None,
                 eos_token=None,
                 pad_token='<pad>',
                 unk_token='<unk>',
                 nb_extra_specials=10,
                 vectors=None,
                 specials_first=True):
        """
        Create a Vocab object from a collections.Counter.
        Arguments:
            counter: collections.Counter object holding the frequencies of
                each value found in the data.
            max_size: The maximum size of the vocabulary, or None for no
                maximum. Default: None.
            min_freq: The minimum frequency needed to include a token in the
                vocabulary. Values less than 1 will be set to 1. Default: 1.
            vectors: One of either the available pretrained vectors
                or custom pretrained vectors (see Vocab.load_vectors);
                or a list of aforementioned vectors
            specials_first: Whether to add special tokens into the vocabulary
                at first. If it is False, they are added into the vocabulary
                at last. Default: True.
        """
        self.unk_token = unk_token
        self.freqs = counter
        counter = counter.copy()
        min_freq = max(min_freq, 1)

        specials = [
            tok for tok in [init_token, eos_token, pad_token, unk_token]
            if tok is not None
        ]
        for i in range(nb_extra_specials):
            specials.append("<SPEC_{}>".format(i))

        self.itos = []
        self.unk_index = None
        if specials_first:
            self.itos = list(specials)
            max_size = None if max_size is None else max_size + len(specials)

        for tok in specials:
            del counter[tok]

        # sort by frequency, then alphabetically
        words_and_frequencies = sorted(counter.items(), key=lambda tup: tup[0])
        words_and_frequencies.sort(key=lambda tup: tup[1], reverse=True)

        for word, freq in words_and_frequencies:
            if freq < min_freq or len(self.itos) == max_size:
                break
            self.itos.append(word)

        if unk_token in specials:
            unk_index = specials.index(unk_token)  # position in list
            # account for ordering of specials, set variable
            self.unk_index = unk_index if specials_first else len(
                self.itos) + unk_index
            self.stoi = defaultdict(self.__default_unk_index)
        else:
            self.stoi = defaultdict()

        if not specials_first:
            self.itos.extend(list(specials))

        # stoi is simply a reverse dict for itos
        self.stoi.update({tok: i for i, tok in enumerate(self.itos)})

        self.vectors = None
        if vectors is not None:
            self.load_vectors(vectors)

    def __default_unk_index(self):
        return self.unk_index

    def extend(self, v, sort=False):
        words = sorted(v.itos) if sort else v.itos
        for w in words:
            if w not in self.stoi:
                self.itos.append(w)
                self.stoi[w] = len(self.itos) - 1

    def load_vectors(self, vectors):
        """
        Arguments:
            vectors: one of or a list containing instantiations
                of Vectors classes.
            Remaining keyword arguments: Passed to the constructor
                of Vectors classes.
        """
        if not isinstance(vectors, list):
            vectors = [vectors]

        tot_dim = sum(v.dim for v in vectors)

        self.vectors = np.zeros((len(self), tot_dim))
        for i, token in enumerate(self.itos):
            start_dim = 0
            for v in vectors:
                end_dim = start_dim + v.dim
                self.vectors[i][start_dim:end_dim] = v[token.strip()]
                start_dim = end_dim
            assert (start_dim == tot_dim)

    def __unk_init(self, x):
        return np.random.uniform(-0.05, 0.05, x.shape)

    def set_vectors(self, stoi, vectors, dim, unk_init=None):
        """
        Set the vectors for the Vocab instance from a collection of Tensors.
        Arguments:
            stoi: A dictionary of string to the index of the associated vector
                in the `vectors` input argument.
            vectors: An indexed iterable (or other structure supporting
                __getitem__) that given an input index, returns a FloatTensor
                representing the vector for the token associated with the
                index. For example, vector[stoi["string"]] should return the
                vector for "string".
            dim: The dimensionality of the vectors.
            unk_init (callback): by default, initialize out-of-vocabulary word
                vectors to zero vectors; can be any function that takes in a
                Tensor and returns a Tensor of the same size.
                Default: lambda x: np.random.uniform(-0.05, 0.05, x.shape)
        """
        if unk_init is None:
            unk_init = self.__unk_init
        self.vectors = np.zeros((len(self), dim))
        for i, token in enumerate(self.itos):
            wv_index = stoi.get(token, None)
            if wv_index is not None:
                self.vectors[i] = vectors[wv_index]
            else:
                self.vectors[i] = unk_init(self.vectors[i])

    def __getitem__(self, token):
        return self.stoi.get(token, self.stoi.get(self.unk_token))

    def __getstate__(self):
        # avoid picking defaultdict
        attrs = dict(self.__dict__)
        # cast to regular dict
        attrs['stoi'] = dict(self.stoi)
        return attrs

    def __setstate__(self, state):
        if state.get("unk_index", None) is None:
            stoi = defaultdict()
        else:
            stoi = defaultdict(self.__default_unk_index)
        stoi.update(state['stoi'])
        state['stoi'] = stoi
        self.__dict__.update(state)

    def __eq__(self, other):
        if not isinstance(other, Vocab):
            return False
        if self.freqs != other.freqs:
            return False
        if self.stoi != other.stoi:
            return False
        if self.itos != other.itos:
            return False
        if (self.vectors != other.vectors).any():
            return False
        return True

    def __len__(self):
        return len(self.itos)


class Vectors(object):
    def __init__(self,
                 file_path,
                 cache=None,
                 prefix=None,
                 unk_init=None,
                 max_vectors=None):
        """
        Arguments:
           file_path: file that contains the vectors
           cache: directory for cached vectors
           prefix: prefix while caching vectors
           unk_init (callback): by default, initialize out-of-vocabulary
               word vectors with a random array of uniform distribution;
               can be any function that takes in a Tensor and
               returns a Tensor of the same size
           max_vectors (int): this can be used to limit the number of
               pre-trained vectors loaded.
               Most pre-trained vector sets are sorted
               in the descending order of word frequency.
               Thus, in situations where the entire set doesn't fit in memory,
               or is not needed for another reason, passing `max_vectors`
               can limit the size of the loaded set.
        """
        cache = '.vector_cache' if cache is None else cache
        self.itos = None
        self.stoi = None
        self.vectors = None
        self.dim = None
        self.unk_init = self.__unk_init if unk_init is None else unk_init
        self.cache(file_path, cache, prefix=prefix, max_vectors=max_vectors)

    def cache(self, file_path, cache, prefix=None, max_vectors=None):
        if os.path.isfile(file_path):
            path = file_path
            if max_vectors:
                file_suffix = '_{}.pkl'.format(max_vectors)
            else:
                file_suffix = '.pkl'
            vec_name = os.path.basename(file_path)
            if prefix:
                vec_name = prefix + vec_name
            path_pt = os.path.join(cache, vec_name) + file_suffix
        else:
            vec_name = file_path
            if prefix:
                vec_name = prefix + vec_name
            path = os.path.join(cache, vec_name)
            if max_vectors:
                file_suffix = '_{}.pkl'.format(max_vectors)
            else:
                file_suffix = '.pkl'
            path_pt = path + file_suffix

        if not os.path.isfile(path_pt):
            if not os.path.isfile(path):
                raise RuntimeError('no vectors found at {}'.format(path))

            print("Loading vectors from {}".format(path))
            ext = os.path.splitext(path)[1][1:]
            if ext == 'gz':
                open_file = gzip.open
            else:
                open_file = open

            vectors_loaded = 0
            with open_file(path, 'rb') as f:
                num_lines, dim = self.__infer_shape(f)
                if not max_vectors or max_vectors > num_lines:
                    max_vectors = num_lines

                itos, vectors, dim = [], np.zeros((max_vectors, dim)), None

                for line in tqdm(f, total=max_vectors):
                    # Explicitly splitting on " " is important, so we don't
                    # get rid of Unicode non-breaking spaces in the vectors.
                    entries = line.rstrip().split(b" ")

                    word, entries = entries[0], entries[1:]
                    if dim is None and len(entries) > 1:
                        dim = len(entries)
                    elif len(entries) == 1:
                        print("Skipping token {} with 1-dimensional "
                              "vector {}; likely a header".format(
                                  word, entries))
                        continue
                    elif dim != len(entries):
                        raise RuntimeError(
                            "Vector for token {} has {} dimensions, but "
                            "previously read vectors have {} dimensions. "
                            "All vectors must have the same number of "
                            "dimensions.".format(word, len(entries), dim))

                    try:
                        if isinstance(word, six.binary_type):
                            word = word.decode('utf-8')
                    except UnicodeDecodeError:
                        print("Skipping non-UTF8 token {}".format(repr(word)))
                        continue

                    vectors[vectors_loaded] = np.asarray(
                        [float(x) for x in entries])
                    vectors_loaded += 1
                    itos.append(word)

                    if vectors_loaded == max_vectors:
                        break

            self.itos = itos
            self.stoi = {word: i for i, word in enumerate(itos)}
            self.vectors = np.asarray(vectors).reshape(-1, dim)
            self.dim = dim
            print('Saving vectors to {}'.format(path_pt))
            if not os.path.exists(cache):
                os.makedirs(cache)
            with open(path_pt, 'wb') as fwb:
                pickle.dump((self.itos, self.stoi, self.vectors, self.dim),
                            fwb)
        else:
            print('Loading vectors from {}'.format(path_pt))
            with open(path_pt, 'rb') as frb:
                self.itos, self.stoi, self.vectors, self.dim = pickle.load(frb)

    def __unk_init(self, x):
        return np.random.uniform(-0.05, 0.05, x.shape)

    def __infer_shape(self, f):
        num_lines, vector_dim = 0, None
        for line in f:
            if vector_dim is None:
                row = line.rstrip().split(b" ")
                vector = row[1:]
                # Assuming word, [vector] format
                if len(vector) > 2:
                    # The header present in some (w2v) formats contains
                    # two elements.
                    vector_dim = len(vector)
                    num_lines += 1  # First element read
            else:
                num_lines += 1
        f.seek(0)
        return num_lines, vector_dim

    def get_vecs_by_tokens(self, tokens, lower_case_backup=False):
        """Look up embedding vectors of tokens.
        Arguments:
            tokens: a token or a list of tokens. if `tokens` is a string,
                returns a 1-D tensor of shape `self.dim`; if `tokens` is a
                list of strings, returns a 2-D tensor of shape=(len(tokens),
                self.dim).
            lower_case_backup : Whether to look up the token in the lower case.
                If False, each token in the original case will be looked up;
                if True, each token in the original case will be looked up
                first, if not found in the keys of the property `stoi`, the
                token in the lower case will be looked up. Default: False.
        Examples:
            >>> examples = ['chip', 'baby', 'Beautiful']
            >>> vec = text.vocab.GloVe(name='6B', dim=50)
            >>> ret = vec.get_vecs_by_tokens(tokens, lower_case_backup=True)
        """
        to_reduce = False

        if not isinstance(tokens, list):
            tokens = [tokens]
            to_reduce = True

        if not lower_case_backup:
            indices = [self[token] for token in tokens]
        else:
            indices = [
                self[token] if token in self.stoi else self[token.lower()]
                for token in tokens
            ]

        vecs = np.asarray(indices)
        return vecs[0] if to_reduce else vecs

    def __len__(self):
        return len(self.vectors)

    def __getitem__(self, token):
        if token in self.stoi:
            return self.vectors[self.stoi[token]]
        else:
            return self.unk_init(np.zeros((self.dim, )))

    def __eq__(self, other):
        if not isinstance(other, Vectors):
            return False
        if self.stoi != other.stoi:
            return False
        if self.itos != other.itos:
            return False
        if self.dim != other.dim:
            return False
        if (self.vectors != other.vectors).any():
            return False
        return True



# -*- coding: utf-8 -*-
from .__field import Field, Fields
from .__entry import Entry
from .__datastorage import DataStorage
from .__vocab import Vocab, Vectors
from . import converter

__version__ = "0.1.19"

__all__ = [
    "Field", "Fields", "Entry", "DataStorage", "Vocab", "Vectors", "converter"
]



# -*- coding: utf-8 -*-
import six


class Entry(object):
    """
    Defines a single training or test entry.
    Stores each column of the entry as an attribute.
    """
    def __init__(self, single_data, fields):
        for (name, field), data in zip(fields.items(), single_data):
            if field is not None:
                if isinstance(data, six.text_type):
                    data = data.rstrip("\n")
                if isinstance(name, tuple):
                    for n, f in zip(name, field):
                        setattr(self, n, f.preprocess(data))
                else:
                    setattr(self, name, field.preprocess(data))
        self.fields = fields

    def get_array(self):
        arr = []
        for name, field in self.fields.items():
            if field is not None:
                arr.append(field.process(getattr(self, name)))
        return arr

    def __str__(self):
        # return str({attr: getattr(self, attr) for attr in self.fields})
        return str([getattr(self, attr) for attr in self.fields])

    def __repr__(self):
        # return str({attr: getattr(self, attr) for attr in self.fields})
        return str([getattr(self, attr) for attr in self.fields])

    def __getitem__(self, attr):
        return getattr(self, attr)

    def __iter__(self):
        return iter(self.fields)

    def __len__(self):
        return len(self.fields)

    def __eq__(self, other):
        if not isinstance(other, Entry):
            return False
        return self.__dict__ == other.__dict__

    def __getstate__(self):
        attrs = dict(self.__dict__)
        return attrs

    def __setstate__(self, state):
        self.__dict__.update(state)



# -*- coding: utf-8 -*-
import pickle

from .__entry import Entry


class DataStorage(object):
    """
    Defines a container to store a dataset composed of entries.

    Attributes:
        fields (dict[str, Field]): Contains the name of each column or field,
            together with the corresponding Field object. Two fields with the
            same Field object will have a shared vocabulary.
        entries (list(Entry)): The entries in this dataset.
    """
    def __init__(self, entries=None):
        self.fields = None
        self.entries = []
        if entries is not None:
            assert isinstance(entries, list), "entries must be list"
            for entry in entries:
                self.add_entry(entry)

    def filter_pred(self, pred):
        """
        pred (callable): Use only entries for which
            filter_pred(entries) is True.
        """
        entries = filter(pred, self.entries)
        self.entries = list(entries)

    def add_entry(self, entry):
        assert isinstance(entry, Entry), "entry must be instance of Entry"
        if self.fields is not None:
            assert entry.fields == self.fields, \
                "fields used in the entry must be the same as used" \
                "in DataStorage"
        else:
            self.fields = entry.fields
        self.entries.append(entry)

    def dump(self, file_name):
        with open(file_name, 'wb') as fwb:
            pickle.dump(self, fwb)

    @classmethod
    def load(cls, file_name):
        with open(file_name, 'rb') as frb:
            cls = pickle.load(frb)
        return cls

    def get_entry(self, i):
        return [getattr(self.entries[i], attr) for attr in self.fields]

    def get_array(self, i):
        return self.entries[i].get_array()

    def sort(self, key=None, reverse=False):
        self.entries.sort(key=key, reverse=reverse)

    def __getitem__(self, i):
        return self.entries[i]

    def __len__(self):
        return len(self.entries)

    def __iter__(self):
        for x in self.entries:
            yield x

    def __getattr__(self, attr):
        if attr in self.fields:
            for x in self.entries:
                yield getattr(x, attr)
        else:
            raise ValueError("{} is not in the fields".format(attr))

    def __getstate__(self):
        attrs = dict(self.__dict__)
        return attrs

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __eq__(self, other):
        if not isinstance(other, DataStorage):
            return False
        return self.__dict__ == other.__dict__



# -*- coding: utf-8 -*-
import torch
import random
import re
from torch._six import container_abcs


class Dataset(torch.utils.data.Dataset):
    def __init__(self, datastorage, key=None, reverse=False):
        self.datastorage = datastorage
        self.fields = datastorage.fields
        if key is not None:
            self.datastorage.sort(key=key, reverse=reverse)

    def __getitem__(self, i):
        return self.datastorage.get_array(i)

    def __len__(self):
        return len(self.datastorage)


class BucketSampler(torch.utils.data.Sampler):
    def __init__(self,
                 sorted_dataset,
                 batch_size=1,
                 shuffle=False,
                 drop_last=False):
        self.dataset = sorted_dataset
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.drop_last = drop_last
        self.all_batches = []
        batch = []
        for idx in range(len(self.dataset)):
            batch.append(idx)
            if len(batch) == self.batch_size:
                self.all_batches.append(batch)
                batch = []
        if len(batch) > 0 and not self.drop_last:
            self.all_batches.append(batch)

    def __iter__(self):
        if self.shuffle:
            random.shuffle(self.all_batches)
        for batch in self.all_batches:
            yield batch

    def __len__(self):
        if self.drop_last:
            return len(self.dataset) // self.batch_size
        else:
            return (len(self.dataset) + self.batch_size - 1) // self.batch_size


class BucketCollate(object):
    def __init__(self, dataset):
        self.fields = dataset.fields

    def __align(self, batch):
        for idx, (name, field) in enumerate(self.fields.items()):
            if field.sequential and not field.fix_length:
                try:
                    pad_idx = field.vocab.stoi[field.pad_token]
                except AttributeError:
                    pad_idx = -1
                max_len = 0
                if field.include_lengths:
                    max_len = max(batch[idx][1]).item()
                    batch[idx][0] = batch[idx][0][:, :max_len]
                else:
                    max_len = max(torch.sum(batch[idx] != pad_idx,
                                            dim=1)).item()
                    batch[idx] = batch[idx][:, :max_len]

    def __call__(self, batch, field=None):
        np_str_obj_array_pattern = re.compile(r'[SaUO]')
        default_collate_err_msg_format = (
            "default_collate: batch must contain tensors, numpy arrays, "
            "numbers, dicts or lists; found {}")

        elem = batch[0]
        elem_type = type(elem)
        if isinstance(elem, torch.Tensor):
            out = None
            if torch.utils.data.get_worker_info() is not None:
                # If we're in a background process, concatenate directly into a
                # shared memory tensor to avoid an extra copy
                numel = sum([x.numel() for x in batch])
                storage = elem.storage()._new_shared(numel)
                out = elem.new(storage)
            return torch.stack(batch, 0, out=out)
        elif elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' \
                and elem_type.__name__ != 'string_':
            elem = batch[0]
            if elem_type.__name__ == 'ndarray':
                # array of string classes and object
                if np_str_obj_array_pattern.search(elem.dtype.str) is not None:
                    raise TypeError(
                        default_collate_err_msg_format.format(elem.dtype))

                return self.__call__([torch.as_tensor(b) for b in batch])
            elif elem.shape == ():  # scalars
                return torch.as_tensor(batch)
        elif isinstance(elem, tuple):  # namedtuple
            transposed = zip(*batch)
            return [self.__call__(samples) for samples in transposed]
        elif isinstance(elem, container_abcs.Sequence):
            transposed = zip(*batch)
            ret = [self.__call__(samples) for samples in transposed]
            self.__align(ret)
            return ret

        raise TypeError(default_collate_err_msg_format.format(elem_type))



# -*- coding: utf-8 -*-
import torch as __torch
from .__dataset import Dataset as __Dataset
from .__dataset import BucketSampler as __BucketSampler
from .__dataset import BucketCollate as __BucketCollate


def to_dataset(datastorage, key=None, reverse=False):
    dataset = __Dataset(datastorage, key=key, reverse=reverse)
    return dataset


def to_dataloader(datastorage,
                  batch_size=1,
                  shuffle=False,
                  num_workers=0,
                  pin_memory=False,
                  drop_last=False,
                  timeout=0,
                  worker_init_fn=None):
    dataset = to_dataset(datastorage)
    dataloader = __torch.utils.data.DataLoader(dataset,
                                               batch_size=batch_size,
                                               shuffle=shuffle,
                                               num_workers=num_workers,
                                               pin_memory=pin_memory,
                                               drop_last=drop_last,
                                               timeout=timeout,
                                               worker_init_fn=worker_init_fn)
    return dataloader


def to_bucketdataloader(datastorage,
                        key=None,
                        reverse=False,
                        batch_size=1,
                        shuffle=False,
                        num_workers=0,
                        pin_memory=False,
                        drop_last=False,
                        timeout=0,
                        worker_init_fn=None):
    dataset = to_dataset(datastorage, key=key, reverse=reverse)
    bucket_sampler = __BucketSampler(dataset,
                                     batch_size=batch_size,
                                     shuffle=shuffle,
                                     drop_last=drop_last)
    bucket_collate = __BucketCollate(dataset)
    dataloader = __torch.utils.data.DataLoader(dataset,
                                               batch_sampler=bucket_sampler,
                                               num_workers=num_workers,
                                               collate_fn=bucket_collate,
                                               pin_memory=pin_memory,
                                               timeout=timeout,
                                               worker_init_fn=worker_init_fn)
    return dataloader


def to_distributeddataloader(datastorage,
                             world_size,
                             rank,
                             batch_size=1,
                             num_workers=0,
                             pin_memory=False,
                             timeout=0,
                             worker_init_fn=None):
    dataset = to_dataset(datastorage)
    distributed_sampler = __torch.utils.data.distributed.DistributedSampler(
        dataset, num_replicas=world_size, rank=rank)
    distributed_loader = __torch.utils.data.DataLoader(
        dataset,
        batch_size=batch_size,
        num_workers=num_workers,
        pin_memory=pin_memory,
        shuffle=False,
        sampler=distributed_sampler,
        timeout=timeout,
        worker_init_fn=worker_init_fn)
    return distributed_loader



# -*- coding: utf-8 -*-
import six
from collections import Counter, OrderedDict
from itertools import chain
import numpy as np
from copy import deepcopy

from .__vocab import Vocab
from .__datastorage import DataStorage


class Field(object):
    """
    Defines a datatype together with instructions for converting to Tensor.

    Field class models common text processing datatypes that can be represented
    by tensors.  It holds a Vocab object that defines the set of possible
    values for elements of the field and their corresponding numerical
    representations.
    The Field object also holds other parameters relating to how a datatype
    should be numericalized, such as a tokenization method and the kind of
    Tensor that should be produced.

    If a Field is shared between two columns in a dataset (e.g., question and
    answer in a QA dataset), then they will have a shared vocabulary.

    Attributes:
        sequential: Whether the datatype represents sequential data. If False,
            no tokenization is applied. Default: True.
        use_vocab: Whether to use a Vocab object. If False, the data in this
            field should already be numerical. Default: True.
        init_token: A token that will be prepended to every example using this
            field, or None for no initial token. Default: None.
        eos_token: A token that will be appended to every example using this
            field, or None for no end-of-sentence token. Default: None.
        fix_length: A fixed length that all examples
            using this field will be padded to
            or None for max sequence lengths. Default: None.
        final_dtype: Data type that represents a batch of examples
            of this kind of data. Default: int.
        preprocessing: The Pipeline that will be applied to examples
            using this field after tokenizing but before numericalizing. Many
            Datasets replace this attribute with a custom preprocessor.
            Default: None.
        postprocessing: A Pipeline that will be applied to examples using
            this field after numericalizing but before the numbers are turned
            into a Tensor. The pipeline function takes the batch as a list, and
            the field's Vocab.
            Default: None.
        lower: Whether to lowercase the text in this field. Default: False.
        include_lengths: Whether to return a tuple of a padded minibatch and
            a list containing the lengths of each examples, or just a padded
            minibatch. Default: False.
        pad_token: The string token used as padding. Default: "<pad>".
        unk_token: The string token used to represent OOV words.
            Default: "<unk>".
        pad_first: Do the padding of the sequence at the beginning.
            Default: False.
        truncate_first: Do the truncating of the sequence at the beginning.
            Default: False
        multiple: If this field contains multiple lines of data.
            Default: False
        stop_words: Tokens to discard during the preprocessing step.
            Default: None
    """
    def __init__(self,
                 sequential=True,
                 use_vocab=True,
                 init_token=None,
                 eos_token=None,
                 fix_length=None,
                 final_dtype=int,
                 preprocessing=None,
                 postprocessing=None,
                 lower=False,
                 include_lengths=False,
                 pad_token="<pad>",
                 unk_token="<unk>",
                 padding=True,
                 pad_first=False,
                 truncate_first=False,
                 multiple=False,
                 stop_words=None):
        self.sequential = sequential
        self.use_vocab = use_vocab
        self.init_token = init_token
        self.eos_token = eos_token
        self.fix_length = fix_length
        self.final_dtype = final_dtype
        self.preprocessing = preprocessing
        self.postprocessing = postprocessing
        self.lower = lower
        self.include_lengths = include_lengths
        self.pad_token = pad_token
        self.unk_token = unk_token
        self.padding = padding if self.sequential and self.pad_token else False
        self.pad_first = pad_first
        self.truncate_first = truncate_first
        self.multiple = multiple
        try:
            self.stop_words = set(
                stop_words) if stop_words is not None else None
        except TypeError:
            raise ValueError("Stop words must be convertible to a set")
        if self.sequential:
            self.max_len = 0
        else:
            self.max_len = None
        if self.use_vocab:
            self.vocab = None

    def __preprocess_single(self, data):
        if self.sequential and isinstance(data, six.text_type):
            data = data.rstrip('\n').split()
        if self.lower:
            data = [d.lowercase() for d in data]
        if self.sequential and self.use_vocab and self.stop_words is not None:
            data = [d for d in data if d not in self.stop_words]
        if self.init_token:
            data.insert(0, self.init_token)
        if self.eos_token:
            data.append(self.eos_token)
        if self.preprocessing is not None:
            data = self.preprocessing(data)
        if self.sequential:
            len_data = len(data) + (self.init_token,
                                    self.eos_token).count(None) - 2
            self.max_len = max(self.max_len, len_data)
        return data

    def preprocess(self, data):
        if self.multiple:
            for idx, d in enumerate(data):
                data[idx] = self.__preprocess_single(d)
        else:
            data = self.__preprocess_single(data)
        return data

    def process(self, x):
        if self.multiple:
            array = []
            for x_line in x:
                padded_line = self.pad(x_line)
                array_line = self.numericalize(padded_line)
                array.append(array_line)

        else:
            padded = self.pad(x)
            array = self.numericalize(padded)
        return array

    def pad(self, inp_x):
        x = deepcopy(inp_x)
        if not self.sequential:
            return x
        if isinstance(self.fix_length, int):
            max_len = self.fix_length + (self.init_token,
                                         self.eos_token).count(None) - 2
        else:
            max_len = self.max_len

        if x:
            if x[0] == self.init_token:
                x.pop(0)
            if x[-1] == self.eos_token:
                x.pop(-1)

        if self.pad_first:
            padded = ([] if not self.padding else[
                    self.pad_token] * max(0, max_len - len(x))) \
                + ([] if self.init_token is None else [self.init_token]) \
                + list(x[-max_len:] if self.truncate_first else x[:max_len]) \
                + ([] if self.eos_token is None else [self.eos_token])
        else:
            padded = ([] if self.init_token is None else [self.init_token]) \
                + list(x[-max_len:] if self.truncate_first else x[:max_len]) \
                + ([] if self.eos_token is None else [self.eos_token]) \
                + ([] if not self.padding else[
                    self.pad_token] * max(0, max_len - len(x)))

        length = len(padded) - max(0, max_len - len(x))

        if self.include_lengths:
            return (padded, length)
        return padded

    def numericalize(self, arr):
        if isinstance(arr, tuple):
            arr, length = arr
            length = np.asarray(length)

        if self.use_vocab:
            if self.sequential:
                arr = [self.vocab.stoi[x] for x in arr]
            else:
                arr = self.vocab.stoi[arr]

            if self.postprocessing is not None:
                arr = self.postprocessing(arr)
        else:
            if self.sequential:
                arr = [
                    self.final_dtype(x)
                    if not isinstance(x, self.final_dtype) else x for x in arr
                ]
            else:
                arr = self.final_dtype(arr) if not isinstance(
                    arr, self.final_dtype) else arr
            if self.postprocessing is not None:
                arr = self.postprocessing(arr)
        var = np.asarray(arr)
        if self.include_lengths:
            return var, length
        return var

    def build_vocab(self, *args, **kwargs):
        """Construct the Vocab object for this field from one or more datasets.
        Arguments:
            *args: Dataset objects or other iterable data
                sources from which to construct the Vocab object that
                represents the set of possible values for this field. If
                a Dataset object is provided, all columns corresponding
                to this field are used; individual columns can also be
                provided directly.
            **kwargs:
                max_size: The maximum size of the vocabulary, or None for no
                    maximum. Default: None.
                min_freq: The minimum frequency needed to include a token in
                    the vocabulary. Values less than 1 will be set to 1.
                    Default: 1.
                nb_extra_specials: number of extra special tokens.
                vectors: One of either the available pretrained vectors
                    or custom pretrained vectors (see Vocab.load_vectors);
                    or a list of aforementioned vectors
                specials_first: Whether to add special tokens into the
                    vocabulary at first. If it is False, they are added into
                    the vocabulary at last. Default: True.
        """
        assert self.use_vocab, "use_vocab is specified to False."
        counter = Counter()
        sources = []
        for arg in args:
            if isinstance(arg, DataStorage):
                sources += [
                    getattr(arg, name) for name, field in arg.fields.items()
                    if field is self
                ]
            else:
                sources.append(arg)
        for data in sources:
            for x in data:
                if not self.sequential:
                    x = [x]
                if self.multiple:
                    x = [j for i in x for j in i]
                try:
                    counter.update(x)
                except TypeError:
                    counter.update(chain.from_iterable(x))
        self.vocab = Vocab(counter,
                           init_token=self.init_token,
                           eos_token=self.eos_token,
                           pad_token=self.pad_token,
                           unk_token=self.unk_token,
                           **kwargs)

    def get_max_len(self):
        max_len = getattr(self, 'max_len') - (self.init_token,
                                              self.eos_token).count(None) + 2
        return max_len

    def __eq__(self, other):
        if not isinstance(other, Field):
            return False
        return self.__dict__ == other.__dict__

    def __getstate__(self):
        attrs = dict(self.__dict__)
        return attrs

    def __setstate__(self, state):
        self.__dict__.update(state)


class Fields(object):
    """
    a collection of fields
    """
    def __init__(self):
        self.__fields = OrderedDict()

    def add_field(self,
                  name,
                  sequential=True,
                  use_vocab=True,
                  init_token=None,
                  eos_token=None,
                  fix_length=None,
                  final_dtype=int,
                  preprocessing=None,
                  postprocessing=None,
                  lower=False,
                  include_lengths=False,
                  pad_token="<pad>",
                  unk_token="<unk>",
                  padding=True,
                  pad_first=False,
                  truncate_first=False,
                  multiple=False,
                  stop_words=None):
        """
        Attributes:
            name: Name of the field.
            sequential: Whether the datatype represents sequential data.
                If False, no tokenization is applied. Default: True.
            use_vocab: Whether to use a Vocab object. If False, the data
                in this field should already be numerical. Default: True.
            init_token: A token that will be prepended to every example
                using this field, or None for no initial token. Default: None.
            eos_token: A token that will be appended to every example using
                this field, or None for no end-of-sentence token.
                Default: None.
            fix_length: A fixed length that all examples
                using this field will be padded to
                or None for max sequence lengths.
                Default: None.
            final_dtype: Data type that represents a batch of examples
                of this kind of data. Default: int.
            preprocessing: The Pipeline that will be applied to examples
                using this field after tokenizing but before numericalizing.
                Many datasets replace this attribute with a custom
                preprocessor. Default: None.
            postprocessing: A Pipeline that will be applied to examples using
                this field after numericalizing but before the numbers are
                turned into a Tensor. The pipeline function takes the batch
                as a list, and the field's Vocab. Default: None.
            lower: Whether to lowercase the text in this field. Default: False.
            include_lengths: Whether to return a tuple of a padded minibatch
                and a list containing the lengths of each examples, or just a
                padded minibatch. Default: False.
            pad_token: The string token used as padding. Default: "<pad>".
            unk_token: The string token used to represent OOV words.
                Default: "<unk>".
            pad_first: Do the padding of the sequence at the beginning.
                Default: False.
            truncate_first: Do the truncating of the sequence at the beginning.
                Default: False
            multiple: If this field contains multiple lines of data.
                Default: False
            stop_words: Tokens to discard during the preprocessing step.
                Default: None
        """
        args = [
            sequential, use_vocab, init_token, eos_token, fix_length,
            final_dtype, preprocessing, postprocessing, lower, include_lengths,
            pad_token, unk_token, padding, pad_first, truncate_first, multiple,
            stop_words
        ]
        self.__fields[name] = Field(*args)

    def del_field(self, name):
        if name in self.__fields:
            self.__fields.pop(name)

    def items(self):
        return list(self.__fields.items())

    def keys(self):
        return list(self.__fields.keys())

    def values(self):
        return list(self.__fields.values())

    def __str__(self):
        return str(self.__fields)

    def __repr__(self):
        return repr(self.__fields)

    def __len__(self):
        return len(self.__fields)

    def __getitem__(self, key):
        return self.__fields[key]

    def __iter__(self):
        return iter(self.keys())

    def __getattr__(self, attr):
        return self.__fields[attr]

    def __getstate__(self):
        attrs = dict(self.__dict__)
        return attrs

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __eq__(self, other):
        if not isinstance(other, Fields):
            return False
        return self.__dict__ == other.__dict__



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import os
import io
import re
from setuptools import setup, find_packages


def read(*names, **kwargs):
    with io.open(os.path.join(os.path.dirname(__file__), *names),
                 encoding=kwargs.get("encoding", "utf8")) as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


VERSION = find_version('preptext', '__init__.py')

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md"), encoding='utf-8') as f:
    long_description = f.read()

setup_info = dict(
    # Metadata
    name='preptext',
    version=VERSION,
    author='Cunliang Kong',
    author_email='cunliang.kong@outlook.com',
    url='https://github.com/styxjedi/preptext',
    description='Utilities for preprocessing texts',
    license='BSD',
    install_requires=['tqdm', 'numpy', 'six', 'torch'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    # Package info
    packages=find_packages(),
    zip_safe=True,
)

setup(**setup_info)
